# -*- coding: UTF-8 -*-
import unittest

from pandasticsearch.queries import Select, Agg


class TestQueries(unittest.TestCase):
    def test_select(self):
        select = Select()
        select._result_dict = {
            'hits': {
                'hits': [
                    {'_source': {'a': 1, 'b': 1}},
                    {'_source': {'a': 2, 'b': 2}},
                    {'_source': {'a': 3, 'b': 3}},
                ]
            }
        }
        select.explain_result()
        print(select)
        print(repr(select))
        print(select.to_pandas())

        self.assertIsNotNone(select.result)
        self.assertEqual(len(select), 3)

    def test_agg_no_buckets(self):
        agg = Agg()
        agg._result_dict = {
            'aggregations': {
                'f1': {'value': 100},
                'f2': {'value': 200},
            }
        }

        agg.explain_result()
        print(agg)
        print(repr(agg))
        print(agg.to_pandas())

        self.assertIsNotNone(agg.result)
        self.assertEqual(len(agg), 1)

    def test_agg_buckets(self):
        agg = Agg()
        agg._result_dict = {
            'aggregations': {
                'agg_key': {
                    'buckets': [
                        {
                            'key': 'a',
                            'f1': {'value': 100},
                            'f2': {'value': 1}
                        },
                        {
                            'key': 'b',
                            'f1': {'value': 200},
                            'f2': {'value': 2}
                        },
                    ]
                }
            }
        }

        agg.explain_result()
        print(agg)
        print(repr(agg))
        print(agg.to_pandas())

        self.assertIsNotNone(agg.result)
        self.assertEqual(len(agg), 2)

    def test_agg_nested_buckets(self):
        agg = Agg()
        agg._result_dict = {
            'aggregations': {
                'agg_key1': {
                    'buckets': [
                        {
                            'key': 'a',
                            'agg_key2': {
                                'buckets': [
                                    {
                                        'key': 'x',
                                        'f1': {'value': 100},
                                        'f2': {'value': 1}
                                    },
                                    {
                                        'key': 'y',
                                        'f1': {'value': 200},
                                        'f2': {'value': 2}
                                    },
                                ]
                            }
                        },
                        {
                            'key': 'b',
                            'agg_key2': {
                                'buckets': [
                                    {
                                        'key': 'x',
                                        'f1': {'value': 300},
                                        'f2': {'value': 3}
                                    },
                                    {
                                        'key': 'y',
                                        'f1': {'value': 400},
                                        'f2': {'value': 4}
                                    },
                                ]
                            }
                        }
                    ]
                }
            }
        }

        agg.explain_result()
        print(agg)
        print(repr(agg))
        print(agg.to_pandas())
        self.assertIsNotNone(agg.result)
        self.assertEqual(len(agg), 4)


if __name__ == '__main__':
    unittest.main()



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os.path
import shlex
import sys

import fixtures
import testtools
import textwrap

from pbr.tests import base
from pbr.tests.test_packaging import CreatePackages
from pbr.tests.test_packaging import Venv

PIPFLAGS = shlex.split(os.environ.get('PIPFLAGS', ''))
PIPVERSION = os.environ.get('PIPVERSION', 'pip')
PBRVERSION = os.environ.get('PBRVERSION', 'pbr')
REPODIR = os.environ.get('REPODIR', '')
WHEELHOUSE = os.environ.get('WHEELHOUSE', '')
PIP_CMD = ['-m', 'pip'] + PIPFLAGS + ['install', '-f', WHEELHOUSE]
PROJECTS = shlex.split(os.environ.get('PROJECTS', ''))
PBR_ROOT = os.path.abspath(os.path.join(__file__, '..', '..', '..'))


def all_projects():
    if not REPODIR:
        return
    # Future: make this path parameterisable.
    excludes = set(['tempest', 'requirements'])
    for name in PROJECTS:
        name = name.strip()
        short_name = name.split('/')[-1]
        try:
            with open(os.path.join(
                    REPODIR, short_name, 'setup.py'), 'rt') as f:
                if 'pbr' not in f.read():
                    continue
        except IOError:
            continue
        if short_name in excludes:
            continue
        yield (short_name, dict(name=name, short_name=short_name))


class TestIntegration(base.BaseTestCase):

    scenarios = list(all_projects())

    def setUp(self):
        # Integration tests need a higher default - big repos can be slow to
        # clone, particularly under guest load.
        env = fixtures.EnvironmentVariable(
            'OS_TEST_TIMEOUT', os.environ.get('OS_TEST_TIMEOUT', '600'))
        with env:
            super(TestIntegration, self).setUp()
        base._config_git()

    @testtools.skipUnless(
        os.environ.get('PBR_INTEGRATION', None) == '1',
        'integration tests not enabled')
    def test_integration(self):
        # Test that we can:
        # - run sdist from the repo in a venv
        # - install the resulting tarball in a new venv
        # - pip install the repo
        # - pip install -e the repo
        # We don't break these into separate tests because we'd need separate
        # source dirs to isolate from side effects of running pip, and the
        # overheads of setup would start to beat the benefits of parallelism.
        self.useFixture(base.CapturedSubprocess(
            'sync-req',
            ['python', 'update.py', os.path.join(REPODIR, self.short_name)],
            cwd=os.path.join(REPODIR, 'requirements')))
        self.useFixture(base.CapturedSubprocess(
            'commit-requirements',
            'git diff --quiet || git commit -amrequirements',
            cwd=os.path.join(REPODIR, self.short_name), shell=True))
        path = os.path.join(
            self.useFixture(fixtures.TempDir()).path, 'project')
        self.useFixture(base.CapturedSubprocess(
            'clone',
            ['git', 'clone', os.path.join(REPODIR, self.short_name), path]))
        venv = self.useFixture(Venv('sdist',
                                    modules=['pip', 'wheel', PBRVERSION],
                                    pip_cmd=PIP_CMD))
        python = venv.python
        self.useFixture(base.CapturedSubprocess(
            'sdist', [python, 'setup.py', 'sdist'], cwd=path))
        venv = self.useFixture(Venv('tarball',
                                    modules=['pip', 'wheel', PBRVERSION],
                                    pip_cmd=PIP_CMD))
        python = venv.python
        filename = os.path.join(
            path, 'dist', os.listdir(os.path.join(path, 'dist'))[0])
        self.useFixture(base.CapturedSubprocess(
            'tarball', [python] + PIP_CMD + [filename]))
        venv = self.useFixture(Venv('install-git',
                                    modules=['pip', 'wheel', PBRVERSION],
                                    pip_cmd=PIP_CMD))
        root = venv.path
        python = venv.python
        self.useFixture(base.CapturedSubprocess(
            'install-git', [python] + PIP_CMD + ['git+file://' + path]))
        if self.short_name == 'nova':
            found = False
            for _, _, filenames in os.walk(root):
                if 'migrate.cfg' in filenames:
                    found = True
            self.assertTrue(found)
        venv = self.useFixture(Venv('install-e',
                                    modules=['pip', 'wheel', PBRVERSION],
                                    pip_cmd=PIP_CMD))
        root = venv.path
        python = venv.python
        self.useFixture(base.CapturedSubprocess(
            'install-e', [python] + PIP_CMD + ['-e', path]))


class TestInstallWithoutPbr(base.BaseTestCase):

    @testtools.skipUnless(
        os.environ.get('PBR_INTEGRATION', None) == '1',
        'integration tests not enabled')
    def test_install_without_pbr(self):
        # Test easy-install of a thing that depends on a thing using pbr
        tempdir = self.useFixture(fixtures.TempDir()).path
        # A directory containing sdists of the things we're going to depend on
        # in using-package.
        dist_dir = os.path.join(tempdir, 'distdir')
        os.mkdir(dist_dir)
        self._run_cmd(sys.executable, ('setup.py', 'sdist', '-d', dist_dir),
                      allow_fail=False, cwd=PBR_ROOT)
        # testpkg - this requires a pbr-using package
        test_pkg_dir = os.path.join(tempdir, 'testpkg')
        os.mkdir(test_pkg_dir)
        pkgs = {
            'pkgTest': {
                'setup.py': textwrap.dedent("""\
                    #!/usr/bin/env python
                    import setuptools
                    setuptools.setup(
                        name = 'pkgTest',
                        tests_require = ['pkgReq'],
                        test_suite='pkgReq'
                    )
                """),
                'setup.cfg': textwrap.dedent("""\
                    [easy_install]
                    find_links = %s
                """ % dist_dir)},
            'pkgReq': {
                'requirements.txt': textwrap.dedent("""\
                    pbr
                """),
                'pkgReq/__init__.py': textwrap.dedent("""\
                    print("FakeTest loaded and ran")
                """)},
        }
        pkg_dirs = self.useFixture(CreatePackages(pkgs)).package_dirs
        test_pkg_dir = pkg_dirs['pkgTest']
        req_pkg_dir = pkg_dirs['pkgReq']

        self._run_cmd(sys.executable, ('setup.py', 'sdist', '-d', dist_dir),
                      allow_fail=False, cwd=req_pkg_dir)
        # A venv to test within
        venv = self.useFixture(Venv('nopbr', ['pip', 'wheel']))
        python = venv.python
        # Run the depending script
        self.useFixture(base.CapturedSubprocess(
            'nopbr', [python] + ['setup.py', 'test'], cwd=test_pkg_dir))


class TestMarkersPip(base.BaseTestCase):

    scenarios = [
        ('pip-1.5', {'modules': ['pip>=1.5,<1.6']}),
        ('pip-6.0', {'modules': ['pip>=6.0,<6.1']}),
        ('pip-latest', {'modules': ['pip']}),
        ('setuptools-EL7', {'modules': ['pip==1.4.1', 'setuptools==0.9.8']}),
        ('setuptools-Trusty', {'modules': ['pip==1.5', 'setuptools==2.2']}),
        ('setuptools-minimum', {'modules': ['pip==1.5', 'setuptools==0.7.2']}),
    ]

    @testtools.skipUnless(
        os.environ.get('PBR_INTEGRATION', None) == '1',
        'integration tests not enabled')
    def test_pip_versions(self):
        pkgs = {
            'test_markers':
                {'requirements.txt': textwrap.dedent("""\
                    pkg_a; python_version=='1.2'
                    pkg_b; python_version!='1.2'
                """)},
            'pkg_a': {},
            'pkg_b': {},
        }
        pkg_dirs = self.useFixture(CreatePackages(pkgs)).package_dirs
        temp_dir = self.useFixture(fixtures.TempDir()).path
        repo_dir = os.path.join(temp_dir, 'repo')
        venv = self.useFixture(Venv('markers'))
        bin_python = venv.python
        os.mkdir(repo_dir)
        for module in self.modules:
            self._run_cmd(
                bin_python,
                ['-m', 'pip', 'install', '--upgrade', module],
                cwd=venv.path, allow_fail=False)
        for pkg in pkg_dirs:
            self._run_cmd(
                bin_python, ['setup.py', 'sdist', '-d', repo_dir],
                cwd=pkg_dirs[pkg], allow_fail=False)
        self._run_cmd(
            bin_python,
            ['-m', 'pip', 'install', '--no-index', '-f', repo_dir,
             'test_markers'],
            cwd=venv.path, allow_fail=False)
        self.assertIn('pkg-b', self._run_cmd(
            bin_python, ['-m', 'pip', 'freeze'], cwd=venv.path,
            allow_fail=False)[0])


class TestLTSSupport(base.BaseTestCase):

    # These versions come from the versions installed from the 'virtualenv'
    # command from the 'python-virtualenv' package.
    scenarios = [
        ('EL7', {'modules': ['pip==1.4.1', 'setuptools==0.9.8'],
                 'py3support': True}),  # And EPEL6
        ('Trusty', {'modules': ['pip==1.5', 'setuptools==2.2'],
                    'py3support': True}),
        ('Jessie', {'modules': ['pip==1.5.6', 'setuptools==5.5.1'],
                    'py3support': True}),
        # Wheezy has pip1.1, which cannot be called with '-m pip'
        # So we'll use a different version of pip here.
        ('WheezyPrecise', {'modules': ['pip==1.4.1', 'setuptools==0.6c11'],
                           'py3support': False})
    ]

    @testtools.skipUnless(
        os.environ.get('PBR_INTEGRATION', None) == '1',
        'integration tests not enabled')
    def test_lts_venv_default_versions(self):
        if (sys.version_info[0] == 3 and not self.py3support):
            self.skipTest('This combination will not install with py3, '
                          'skipping test')
        venv = self.useFixture(Venv('setuptools', modules=self.modules))
        bin_python = venv.python
        pbr = 'file://%s#egg=pbr' % PBR_ROOT
        # Installing PBR is a reasonable indication that we are not broken on
        # this particular combination of setuptools and pip.
        self._run_cmd(bin_python, ['-m', 'pip', 'install', pbr],
                      cwd=venv.path, allow_fail=False)



# -*- coding: UTF-8 -*-
import unittest

from pandasticsearch.filters import *


class TestFilters(unittest.TestCase):
    def test_leaves(self):
        self.assertEqual((Dim('a') >= 2).build(), {"range": {"a": {"gte": 2}}})
        self.assertEqual((Dim('a') <= 2).build(), {"range": {"a": {"lte": 2}}})
        self.assertEqual((Dim('a') < 2).build(), {"range": {"a": {"lt": 2}}})
        self.assertEqual((Dim('a') == 2).build(), {"term": {"a": 2}})
        self.assertEqual((Dim('a') != 2).build()['bool'], {"must_not": {"term": {"a": 2}}})
        self.assertEqual((Dim('a') > 2).build(), {"range": {"a": {"gt": 2}}})

    def test_and(self):
        self.assertEqual(
            ((Dim('a') >= 2) & (Dim('b') < 3)).build()['bool'],
            {
                'must': [
                    {'range': {'a': {'gte': 2}}},
                    {'range': {'b': {'lt': 3}}}]
            })

        self.assertEqual(
            ((Dim('a') >= 2) & (Dim('b') < 3) & (Dim('c') == 4)).build()['bool'],
            {
                'must': [
                    {'range': {'a': {'gte': 2}}},
                    {'range': {'b': {'lt': 3}}},
                    {'term': {'c': 4}}]
            })

        self.assertEqual(
            ((Dim('a') >= 2) & ((Dim('b') < 3) & (Dim('c') == 4))).build()['bool'],
            {
                'must': [
                    {'range': {'b': {'lt': 3}}},
                    {'term': {'c': 4}},
                    {'range': {'a': {'gte': 2}}}]
            })

    def test_or(self):
        self.assertEqual(
            ((Dim('a') >= 2) | (Dim('b') < 3)).build()['bool'],
            {
                'should': [
                    {'range': {'a': {'gte': 2}}},
                    {'range': {'b': {'lt': 3}}}]
            })

        self.assertEqual(
            ((Dim('a') >= 2) | (Dim('b') < 3) | (Dim('c') == 4)).build()['bool'],
            {
                'should': [
                    {'range': {'a': {'gte': 2}}},
                    {'range': {'b': {'lt': 3}}},
                    {'term': {'c': 4}}]
            })

        self.assertEqual(
            ((Dim('a') >= 2) | ((Dim('b') < 3) | (Dim('c') == 4))).build()['bool'],
            {
                'should': [
                    {'range': {'b': {'lt': 3}}},
                    {'term': {'c': 4}},
                    {'range': {'a': {'gte': 2}}}]
            })

    def test_not(self):
        self.assertEqual(
            (~(Dim('a') >= 2)).build()['bool'],
            {
                'must_not':
                    {'range': {'a': {'gte': 2}}}})

    def test_not_not(self):
        exp = Dim('a') >= 2
        print((~exp).build())
        print((~~exp).build())
        self.assertEqual(
            (~~exp).build()['bool'],
            {
                'must_not': {
                    'bool': {
                        'must_not':
                            {'range': {'a': {'gte': 2}}}}}})

    def test_not_and(self):
        exp = (Dim('a') >= 2) & (Dim('b') < 3)
        self.assertEqual(
            (~exp).build()['bool'],
            {
                'must_not': {
                    'bool': {
                        'must': [
                            {'range': {'a': {'gte': 2}}},
                            {'range': {'b': {'lt': 3}}}]
                    }
                }
            })

    def test_and_or(self):
        exp = (Dim('b') < 3) | (Dim('c') == 4)
        actual = (Dim('a') >= 2) & exp
        print(actual.debug_string())

        self.assertEqual(
            actual.build()['bool'],
            {
                'must': [
                    {'range': {'a': {'gte': 2}}},
                    {
                        'bool': {
                            'should': [
                                {'range': {'b': {'lt': 3}}},
                                {'term': {'c': 4}}
                            ]
                        }
                    }]
            })


if __name__ == '__main__':
    unittest.main()



# -*- coding: UTF-8 -*-
import unittest
from mock import patch, Mock

from pandasticsearch.api import Pandasticsearch


class TestAPI(unittest.TestCase):
    @patch('pandasticsearch.clients.urllib.request.urlopen')
    def test_pandasticsearch_top(self, mock_urlopen):
        response = Mock()
        response.read.return_value = """{"hits" : {"hits": [{"_source": {}}] }}""".encode("utf-8")
        mock_urlopen.return_value = response
        ps = Pandasticsearch("http://localhost:9200", 'xxx')

        top = ps.top()
        print(repr(top))
        self.assertEqual(repr(top), 'Select: 1 rows')


if __name__ == '__main__':
    unittest.main()



from pandasticsearch.clients import RestClient
from pandasticsearch.queries import Agg, Select
from pandasticsearch.filters import Filter
from pandasticsearch.utils.metric import metric_agg

import json


class Pandasticsearch(object):
    def __init__(self, url, index, type=None):
        if type is None:
            endpoint = index + '/_search'
        else:
            endpoint = index + '/' + type + '/_search'
        self._client = RestClient(url, endpoint)
        self._filter = None
        self._last_query = None

    def where(self, filter=None):
        assert isinstance(filter, Filter)
        self._filter = filter.build()
        return self

    def top(self, size=10):
        self._last_query = Pandasticsearch._build_query(filter=self._filter, size=size)
        return self._client.execute(self._last_query, Select())

    def aggregate(self, agg_func):
        self._last_query = Pandasticsearch._build_query(agg=agg_func, filter=self._filter, size=0)
        return self._client.execute(self._last_query, Agg())

    @staticmethod
    def _build_query(agg=None, filter=None, size=None):
        query = {}
        if agg is not None:
            query['aggs'] = agg
        if size is not None:
            query['size'] = size
        if filter is not None:
            query['query'] = {'filtered': {'filter': filter}}
        else:
            query['query'] = {"match_all": {}}
        return query

    def debug_string(self, indent=4):
        return json.dumps(self._last_query, indent=indent)

    def distinct_count(self, field):
        """
        Calculates an approximate count of distinct values.
        :param str field: The field to be performed metric-aggregation on
        :return: an Agg object containing the aggregated value
        :rtype: Agg
        """
        agg = metric_agg('cardinality', field)
        self._last_query = Pandasticsearch._build_query(agg=agg, filter=self._filter, size=0)
        return self._client.execute(self._last_query, Agg())

    def value_count(self, field):
        """
        Counts the number of values that are extracted from the aggregated documents.
        :param str field: The field to be performed metric-aggregation on
        :return: an Agg object containing the aggregated value
        :rtype: Agg
        """
        agg = metric_agg('value_count', field)
        self._last_query = Pandasticsearch._build_query(agg, filter=self._filter, size=0)
        return self._client.execute(self._last_query, Agg())

    def percentiles(self, field, percents=None):
        """
        Calculates one or more percentiles over numeric values extracted from the aggregated documents.
        :param str field: The field to be performed metric-aggregation on
        :return: an Agg object containing the aggregated value
        :rtype: Agg
        """
        agg = metric_agg('percentiles', field, params={'percents': percents})
        self._last_query = Pandasticsearch._build_query(agg=agg, filter=self._filter, size=0)
        return self._client.execute(self._last_query, Agg())

    def percentile_ranks(self, field, values=None):
        """
        Calculates one or more percentile ranks over numeric values extracted from the aggregated documents.
        :param str field: The field to be performed metric-aggregation on
        :return: an Agg object containing the aggregated value
        :rtype: Agg
        """
        agg = metric_agg('percentile_ranks', field, params={'values': values})
        self._last_query = Pandasticsearch._build_query(agg=agg, filter=self._filter, size=0)
        return self._client.execute(self._last_query, Agg())



import json
import sys
from six.moves import urllib

from pandasticsearch.queries import Query
from pandasticsearch.errors import ServerDefinedException


class RestClient(object):
    """
    RestClient talks to Elasticsearch cluster through native RESTful API.
    Returns Query objects that can be used to export to pandas.DataFrame objects for subsequent analysis.
    :param str url: URL of Broker node in the Elasticsearch cluster
    :param str endpoint: Endpoint that Broker listens for queries on

    :Example:
    >>> from pandasticsearch.clients import RestClient
    >>> client = RestClient('http://localhost:9200', 'index/type/_search')
    >>> query = client.execute("query":{"match_all":{}}})
    >>> print query
    """

    def __init__(self, url, endpoint=''):
        self.url = url
        self.endpoint = endpoint

    def execute(self, data, query=Query()):
        res = self._post(data=data)
        query.parse_json(res)
        query.explain_result()
        return query

    def _prepare_url(self):
        if self.url.endswith('/'):
            url = self.url + self.endpoint
        else:
            url = self.url + '/' + self.endpoint
        return url

    def _post(self, **kwargs):
        try:
            url = self._prepare_url()
            data = kwargs.get('data', None)
            params = kwargs.get('params', None)

            if params is not None:
                url = '{0}?{1}'.format(url, urllib.parse.urlencode(params))

            req = urllib.request.Request(url=url, data=json.dumps(data).encode('utf-8'),
                                         headers={'Content-Type': 'application/json'})
            res = urllib.request.urlopen(req)
            data = res.read().decode("utf-8")
            res.close()
        except urllib.error.HTTPError:
            _, e, _ = sys.exc_info()
            reason = None
            if e.code != 200:
                try:
                    reason = json.loads(e.read().decode("utf-8"))
                except (ValueError, AttributeError, KeyError):
                    pass
                else:
                    reason = reason.get('error', None)

            raise ServerDefinedException(reason)
        else:
            return data


class SqlClient(RestClient):
    """
    SqlClient talks to Elasticsearch cluster through SQL.
    Returns Query objects that can be used to export to pandas.DataFrame objects for subsequent analysis.
    :param str url: URL of Broker node in the Elasticsearch cluster
    :param str endpoint: Endpoint that Broker listens for queries on

    :Example:
    >>> from pandasticsearch.clients import SqlClient
    >>> client = SqlClient('http://localhost:9200')
    >>> query = client.execute('select * from table_name')
    >>> print query, query.json
    """

    def __init__(self, url):
        super(SqlClient, self).__init__(url, '_sql')

    def execute(self, sql, query=Query()):
        res = self._post(params={'sql': sql})
        query.parse_json(res)
        query.explain_result()
        return query



import json


# Bool filter builder
class Filter(object):
    def __init__(self, *args):
        self._filter = None

    def __and__(self, x):
        # Combine results
        if isinstance(self, And):
            self.subtree['must'].append(x.subtree)
            return self
        elif isinstance(x, And):
            x.subtree['must'].append(self.subtree)
            return x
        return And(self, x)

    def __or__(self, x):
        # Combine results
        if isinstance(self, Or):
            self.subtree['should'].append(x.subtree)
            return self
        elif isinstance(x, Or):
            x.subtree['should'].append(self.subtree)
            return x
        return Or(self, x)

    def __invert__(self):
        return Not(self)

    @property
    def subtree(self):
        if 'bool' in self._filter:
            return self._filter['bool']
        else:
            return self._filter

    def build(self):
        return self._filter

    def debug_string(self, indent=4):
        return json.dumps(self._filter, indent=indent)


# Binary operator
class And(Filter):
    def __init__(self, *args):
        [isinstance(x, Filter) for x in args]
        super(And, self).__init__()
        self._filter = {'bool': {'must': [x.build() for x in args]}}


class Or(Filter):
    def __init__(self, *args):
        [isinstance(x, Filter) for x in args]
        super(Or, self).__init__()
        self._filter = {'bool': {'should': [x.build() for x in args]}}


class Not(Filter):
    def __init__(self, x):
        assert isinstance(x, Filter)
        super(Not, self).__init__()
        self._filter = {'bool': {'must_not': x.build()}}


# Leaves
class GreaterEqual(Filter):
    def __init__(self, dim, value):
        super(GreaterEqual, self).__init__()
        self._filter = {'range': {dim: {'gte': value}}}


class Greater(Filter):
    def __init__(self, dim, value):
        super(Greater, self).__init__()
        self._filter = {'range': {dim: {'gt': value}}}


class LessEqual(Filter):
    def __init__(self, dim, value):
        super(LessEqual, self).__init__()
        self._filter = {'range': {dim: {'lte': value}}}


class Less(Filter):
    def __init__(self, dim, value):
        super(Less, self).__init__()
        self._filter = {'range': {dim: {'lt': value}}}


class Equal(Filter):
    def __init__(self, dim, value):
        super(Equal, self).__init__()
        self._filter = {'term': {dim: value}}


class IsIn(Filter):
    def __init__(self, dim, value):
        super(IsIn, self).__init__()
        self._filter = {'terms': {dim: value}}


class Dim(object):
    def __init__(self, dim):
        self._dim = dim

    def __eq__(self, other):
        return Equal(dim=self._dim, value=other)

    def __ne__(self, other):
        return ~Equal(dim=self._dim, value=other)

    def __gt__(self, other):
        return Greater(dim=self._dim, value=other)

    def __lt__(self, other):
        return Less(dim=self._dim, value=other)

    def __ge__(self, other):
        return GreaterEqual(dim=self._dim, value=other)

    def __le__(self, other):
        return LessEqual(dim=self._dim, value=other)

    def isin(self, other):
        return IsIn(dim=self._dim, value=other)



def metric_agg(agg_type, field, metric_rename=None, params=None):
    if agg_type in ('value_count', 'cardinality', 'percentiles', 'percentile_ranks', 'avg', 'max', 'min'):
        pass
    else:
        raise NotImplementedError('type={0} is not supported for metric agg'.format(agg_type))

    if metric_rename is None:
        metric_rename = '{0}({1})'.format(agg_type, field)

    agg_field = dict()
    agg_field['field'] = field
    if params is not None:
        agg_field.update(params)

    return {metric_rename: {agg_type: agg_field}}


def avg(field):
    agg_func = metric_agg('avg', field)
    return agg_func


def min(field):
    agg_func = metric_agg('min', field)
    return agg_func


def max(field):
    agg_func = metric_agg('max', field)
    return agg_func


# Copyright 2011 OpenStack LLC.
# Copyright 2012-2013 Hewlett-Packard Development Company, L.P.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from __future__ import unicode_literals

import distutils.errors
from distutils import log
import errno
import io
import os
import re
import subprocess
import time

import pkg_resources

from pbr import options
from pbr import version


def _run_shell_command(cmd, throw_on_error=False, buffer=True, env=None):
    if buffer:
        out_location = subprocess.PIPE
        err_location = subprocess.PIPE
    else:
        out_location = None
        err_location = None

    newenv = os.environ.copy()
    if env:
        newenv.update(env)

    output = subprocess.Popen(cmd,
                              stdout=out_location,
                              stderr=err_location,
                              env=newenv)
    out = output.communicate()
    if output.returncode and throw_on_error:
        raise distutils.errors.DistutilsError(
            "%s returned %d" % (cmd, output.returncode))
    if len(out) == 0 or not out[0] or not out[0].strip():
        return ''
    # Since we don't control the history, and forcing users to rebase arbitrary
    # history to fix utf8 issues is harsh, decode with replace.
    return out[0].strip().decode('utf-8', 'replace')


def _run_git_command(cmd, git_dir, **kwargs):
    if not isinstance(cmd, (list, tuple)):
        cmd = [cmd]
    return _run_shell_command(
        ['git', '--git-dir=%s' % git_dir] + cmd, **kwargs)


def _get_git_directory():
    try:
        return _run_shell_command(['git', 'rev-parse', '--git-dir'])
    except OSError as e:
        if e.errno == errno.ENOENT:
            # git not installed.
            return ''
        raise


def _git_is_installed():
    try:
        # We cannot use 'which git' as it may not be available
        # in some distributions, So just try 'git --version'
        # to see if we run into trouble
        _run_shell_command(['git', '--version'])
    except OSError:
        return False
    return True


def _get_highest_tag(tags):
    """Find the highest tag from a list.

    Pass in a list of tag strings and this will return the highest
    (latest) as sorted by the pkg_resources version parser.
    """
    return max(tags, key=pkg_resources.parse_version)


def _find_git_files(dirname='', git_dir=None):
    """Behave like a file finder entrypoint plugin.

    We don't actually use the entrypoints system for this because it runs
    at absurd times. We only want to do this when we are building an sdist.
    """
    file_list = []
    if git_dir is None:
        git_dir = _run_git_functions()
    if git_dir:
        log.info("[pbr] In git context, generating filelist from git")
        file_list = _run_git_command(['ls-files', '-z'], git_dir)
        # Users can fix utf8 issues locally with a single commit, so we are
        # strict here.
        file_list = file_list.split(b'\x00'.decode('utf-8'))
    return [f for f in file_list if f]


def _get_raw_tag_info(git_dir):
    describe = _run_git_command(['describe', '--always'], git_dir)
    if "-" in describe:
        return describe.rsplit("-", 2)[-2]
    if "." in describe:
        return 0
    return None


def get_is_release(git_dir):
    return _get_raw_tag_info(git_dir) == 0


def _run_git_functions():
    git_dir = None
    if _git_is_installed():
        git_dir = _get_git_directory()
    return git_dir or None


def get_git_short_sha(git_dir=None):
    """Return the short sha for this repo, if it exists."""
    if not git_dir:
        git_dir = _run_git_functions()
    if git_dir:
        return _run_git_command(
            ['log', '-n1', '--pretty=format:%h'], git_dir)
    return None


def _iter_changelog(changelog):
    """Convert a oneline log iterator to formatted strings.

    :param changelog: An iterator of one line log entries like
        that given by _iter_log_oneline.
    :return: An iterator over (release, formatted changelog) tuples.
    """
    first_line = True
    current_release = None
    yield current_release, "CHANGES\n=======\n\n"
    for hash, tags, msg in changelog:
        if tags:
            current_release = _get_highest_tag(tags)
            underline = len(current_release) * '-'
            if not first_line:
                yield current_release, '\n'
            yield current_release, (
                "%(tag)s\n%(underline)s\n\n" %
                dict(tag=current_release, underline=underline))

        if not msg.startswith("Merge "):
            if msg.endswith("."):
                msg = msg[:-1]
            yield current_release, "* %(msg)s\n" % dict(msg=msg)
        first_line = False


def _iter_log_oneline(git_dir=None):
    """Iterate over --oneline log entries if possible.

    This parses the output into a structured form but does not apply
    presentation logic to the output - making it suitable for different
    uses.

    :return: An iterator of (hash, tags_set, 1st_line) tuples, or None if
        changelog generation is disabled / not available.
    """
    if git_dir is None:
        git_dir = _get_git_directory()
    if not git_dir:
        return []
    return _iter_log_inner(git_dir)


def _is_valid_version(candidate):
    try:
        version.SemanticVersion.from_pip_string(candidate)
        return True
    except ValueError:
        return False


def _iter_log_inner(git_dir):
    """Iterate over --oneline log entries.

    This parses the output intro a structured form but does not apply
    presentation logic to the output - making it suitable for different
    uses.

    :return: An iterator of (hash, tags_set, 1st_line) tuples.
    """
    log.info('[pbr] Generating ChangeLog')
    log_cmd = ['log', '--format=%h%x00%s%x00%d']
    changelog = _run_git_command(log_cmd, git_dir)
    for line in changelog.split('\n'):
        line_parts = line.split('\x00')
        if len(line_parts) != 3:
            continue
        sha, msg, refname = line_parts
        tags = set()

        # refname can be:
        #  <empty>
        #  HEAD, tag: 1.4.0, origin/master, master
        #  tag: 1.3.4
        if "tag:" in refname:
            refname = refname.strip()[1:-1]  # remove wrapping ()'s
            # If we start with "tag: 1.2b1, tag: 1.2"
            # The first split gives us "['', '1.2b1, ', '1.2']"
            # Which is why we do the second split below on the comma
            for tag_string in refname.split("tag: ")[1:]:
                # git tag does not allow : or " " in tag names, so we split
                # on ", " which is the separator between elements
                candidate = tag_string.split(", ")[0]
                if _is_valid_version(candidate):
                    tags.add(candidate)

        yield sha, tags, msg


def write_git_changelog(git_dir=None, dest_dir=os.path.curdir,
                        option_dict=None, changelog=None):
    """Write a changelog based on the git changelog."""
    start = time.time()
    if not option_dict:
        option_dict = {}
    should_skip = options.get_boolean_option(option_dict, 'skip_changelog',
                                             'SKIP_WRITE_GIT_CHANGELOG')
    if should_skip:
        return
    if not changelog:
        changelog = _iter_log_oneline(git_dir=git_dir)
        if changelog:
            changelog = _iter_changelog(changelog)
    if not changelog:
        return
    new_changelog = os.path.join(dest_dir, 'ChangeLog')
    # If there's already a ChangeLog and it's not writable, just use it
    if (os.path.exists(new_changelog)
            and not os.access(new_changelog, os.W_OK)):
        log.info('[pbr] ChangeLog not written (file already'
                 ' exists and it is not writeable)')
        return
    log.info('[pbr] Writing ChangeLog')
    with io.open(new_changelog, "w", encoding="utf-8") as changelog_file:
        for release, content in changelog:
            changelog_file.write(content)
    stop = time.time()
    log.info('[pbr] ChangeLog complete (%0.1fs)' % (stop - start))


def generate_authors(git_dir=None, dest_dir='.', option_dict=dict()):
    """Create AUTHORS file using git commits."""
    should_skip = options.get_boolean_option(option_dict, 'skip_authors',
                                             'SKIP_GENERATE_AUTHORS')
    if should_skip:
        return
    start = time.time()
    old_authors = os.path.join(dest_dir, 'AUTHORS.in')
    new_authors = os.path.join(dest_dir, 'AUTHORS')
    # If there's already an AUTHORS file and it's not writable, just use it
    if (os.path.exists(new_authors)
            and not os.access(new_authors, os.W_OK)):
        return
    log.info('[pbr] Generating AUTHORS')
    ignore_emails = '(jenkins@review|infra@lists|jenkins@openstack)'
    if git_dir is None:
        git_dir = _get_git_directory()
    if git_dir:
        authors = []

        # don't include jenkins email address in AUTHORS file
        git_log_cmd = ['log', '--format=%aN <%aE>']
        authors += _run_git_command(git_log_cmd, git_dir).split('\n')
        authors = [a for a in authors if not re.search(ignore_emails, a)]

        # get all co-authors from commit messages
        co_authors_out = _run_git_command('log', git_dir)
        co_authors = re.findall('Co-authored-by:.+', co_authors_out,
                                re.MULTILINE)
        co_authors = [signed.split(":", 1)[1].strip()
                      for signed in co_authors if signed]

        authors += co_authors
        authors = sorted(set(authors))

        with open(new_authors, 'wb') as new_authors_fh:
            if os.path.exists(old_authors):
                with open(old_authors, "rb") as old_authors_fh:
                    new_authors_fh.write(old_authors_fh.read())
            new_authors_fh.write(('\n'.join(authors) + '\n')
                                 .encode('utf-8'))
    stop = time.time()
    log.info('[pbr] AUTHORS complete (%0.1fs)' % (stop - start))



import collections
import json
import pandas

from pandasticsearch.errors import NoSuchDependencyException


class Query(collections.MutableSequence):
    """
    Query objects are produced by Elasticsearch clients and can be used for exporting query results into
    pandas.DataFrame objects for subsequent analysis.
    """

    def __init__(self):
        super(Query, self).__init__()
        self._values = None
        self._result_dict = None

    def parse_json(self, json_str):
        res = json.loads(json_str)
        self._result_dict = res

    def explain_result(self, result=None):
        if result is not None:
            assert isinstance(result, dict)
            self._result_dict = result

    def to_pandas(self):
        """
        Export the current query result to a Pandas DataFrame object.
        :return: The DataFrame representing the query result
        :rtype: DataFrame
        :raise: NotImplementedError
        """
        raise NotImplementedError('implemented in subclass')

    def pretty(self, indent=None):
        """
        Prettify the result (requires pygements package)
        :return: The formatted string
        :rtype: string
        :raise: ImportError
        """
        try:
            from pygments import highlight, lexers, formatters
            indented = json.dumps(self._result_dict, sort_keys=True, separators=(',', ': '), indent=indent,
                                  ensure_ascii=False)
            return highlight(indented, lexers.JsonLexer(), formatters.TerminalFormatter())
        except ImportError:
            raise NoSuchDependencyException('pretty() method requires pygements package')

    @property
    def result(self):
        return self._values

    @property
    def json(self):
        """
        Gets the original JSON representation returned by Elasticsearch REST API
        :return: The JSON string indicating the query result
        :rtype: string
        """
        return json.dumps(self._result_dict)

    def insert(self, index, value):
        self._values.insert(index, value)

    def append(self, value):
        self._values.append(value)

    def __str__(self):
        return str(self._values)

    def __len__(self):
        return len(self._values)

    def __delitem__(self, index):
        del self._values[index]

    def __setitem__(self, index, value):
        self._values[index] = value

    def __getitem__(self, index):
        return self._values[index]


class Select(Query):
    def __init__(self):
        super(Select, self).__init__()
        self._values = []

    def __repr__(self):
        return 'Select: {0} rows'.format(len(self._values))

    def explain_result(self, result=None):
        super(Select, self).explain_result(result)
        self._values = [hit['_source'] for hit in self._result_dict['hits']['hits']]

    def to_pandas(self):
        if self._values:
            df = pandas.DataFrame(data=self._values)
            return df


class Agg(Query):
    def __init__(self):
        super(Agg, self).__init__()
        self._index_names = []
        self._indexes = []
        self._values = []

    def __repr__(self):
        return 'Agg: {0} rows'.format(len(self._values))

    def explain_result(self, result=None):
        super(Agg, self).explain_result(result)
        tuples = list(Agg._process_agg(self._result_dict['aggregations']))
        assert len(tuples) > 0
        self._index_names = list(tuples[0][0])
        for t in tuples:
            _, index, row = t
            self.append(row)
            if len(index) > 0:
                self._indexes.append(index)

    def to_pandas(self):
        if self._values:
            if len(self._indexes) > 0:
                index = pandas.MultiIndex.from_tuples(self._indexes, names=self._index_names)
                df = pandas.DataFrame(data=self._values, index=index)
            else:
                df = pandas.DataFrame(data=self._values)
            return df

    @classmethod
    def _process_agg(cls, bucket, indexes=(), names=()):
        """
        Recursively extract agg values
        :param bucket: a bucket contains either sub-buckets or a bunch of aggregated values
        :return: a list of tuples: (index_name, index_tuple, row)
        """
        # for each agg, yield a row
        row = {}
        for k, v in bucket.items():
            if 'buckets' in v:
                for sub_bucket in v['buckets']:
                    for x in Agg._process_agg(sub_bucket, indexes + (sub_bucket['key'],), names + (k,)):
                        yield x
            elif 'value' in v:
                row[k] = v['value']
            elif 'values' in v:  # percentiles
                row = v['values']
            else:
                pass
        if len(row) > 0:
            yield (names, indexes, row)




from setuptools import setup

install_requires = [
    "six >= 1.9.0",
]

extras_require = {
    "official": ["elasticsearch"],
}

setup(
    name='pandasticsearch',
    version='0.0.9',
    author='onesuper',
    author_email='onesuperclark@gmail.com',
    packages=['pandasticsearch', 'pandasticsearch.utils'],
    url='http://pypi.python.org/pypi/pandasticsearch/',
    license='MIT',
    description='A Pandastic Elasticsearch client for data analyzing.',
    install_requires=install_requires,
    extras_require=extras_require,
    test_suite='nose.collector',
    tests_require=['pandas', 'nose', 'mock'],
)
